<?php

declare(strict_types=1);

namespace Drupal\ui_patterns\Plugin\UiPatterns\Source;

use Drupal\Component\Utility\Html;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Security\TrustedCallbackInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\ui_patterns\Attribute\Source;
use Drupal\ui_patterns\SourcePluginPropValue;
use Drupal\ui_patterns\UnicodePatternValidatorTrait;

/**
 * Plugin implementation of the source.
 */
#[Source(
  id: 'attributes',
  label: new TranslatableMarkup('Attributes'),
  description: new TranslatableMarkup('Textfield with double-quoted values or a space-separated list of HTML classes.'),
  prop_types: ['attributes'],
)]
class AttributesWidget extends SourcePluginPropValue implements TrustedCallbackInterface {

  use UnicodePatternValidatorTrait;

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks() {
    return ['validateUnicodePattern'];
  }

  /**
   * {@inheritdoc}
   */
  public function getPropValue(): mixed {
    // In UI Patterns Settings, we built the Attribute object here. It is not
    // possible anymore because SDC will not validate it against the prop
    // type schema.
    return static::convertValueToAttributesMapping(parent::getPropValue());
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state): array {
    $form = parent::settingsForm($form, $form_state);
    // Attributes are associative arrays, but this source plugin is storing
    // them as string in config.
    // It would be better to use something else than a textfield one day.
    $value = $this->getSetting('value') ?? '';
    $form['value'] = [
      '#type' => 'textfield',
      '#default_value' => $value,
    ];
    if (static::isValueForAttributes($value)) {
      $form['value']['#pattern_unicode'] = static::buildAttributesRegexPattern();
      $form['value']['#pattern_error'] = t('Format is not HTML attributes with double-quoted values.');
    }
    else {
      $form['value']['#pattern_unicode'] = static::buildClassRegexPattern();
      $form['value']['#pattern_error'] = t('Not a valid space-separated list of HTML classes.');
    }
    // To allow form errors to be displayed correctly.
    $form['value']['#title'] = '';
    $form['value']['#placeholder'] = 'class="hidden" title="Lorem ipsum"';
    $form['value']['#description'] = $this->t("HTML attributes with double-quoted values or a space-separated list of HTML classes.");
    $this->addRequired($form['value']);
    $form['value']['#element_validate'][] = [static::class, 'validateUnicodePattern'];
    return $form;
  }

  /**
   * Build regular expression pattern.
   *
   * See https://html.spec.whatwg.org/#attributes-2
   */
  protected static function buildAttributesRegexPattern(): string {
    // Attribute names are a mix of ASCII lower and upper alphas.
    $attr_name = "[a-zA-Z\-]+";
    // Discard double quotes which are used for delimiting.
    $double_quoted_value = '[^"]*';
    $space = "\s*";
    $attr = sprintf("%s=\"%s\"%s", $attr_name, $double_quoted_value, $space);
    // Start and end delimiters are not expected here, they will be added:
    // - by \Drupal\Core\Render\Element\FormElementBase::validatePattern for
    //   server side validation
    // - in the HTML5 pattern attribute, for client side validation
    // https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes/pattern
    return $space . "(" . $attr . ")*";
  }

  /**
   * Build regular expression pattern.
   */
  protected static function buildClassRegexPattern(): string {
    // Each classname cannot start with a hyphen followed by a digit or a digit.
    $class_name = "(?!(?:\\d|[-]\\d))[\\S]+";
    // Pattern for valid CSS class names.
    return "\\s*(" . $class_name . "(\\s+" . $class_name . ")*)?\\s*";
  }

  /**
   * Check if the value is for attributes.
   *
   * @param string $value
   *   The value to check.
   *
   * @return bool
   *   TRUE if the value is for attributes, FALSE otherwise.
   */
  protected static function isValueForAttributes(string $value): bool {
    return str_contains($value, '=');
  }

  /**
   * Convert a string to an attribute mapping.
   *
   * @param mixed $value
   *   The string to convert.
   *
   * @return array
   *   Attributes mapping.
   */
  protected static function convertValueToAttributesMapping(mixed $value): array {
    if (empty($value) || !is_string($value)) {
      return [];
    }
    if (static::isValueForAttributes($value)) {
      return static::convertStringToAttributesMapping($value);
    }
    return static::convertStringToAttributesMapping(sprintf('class="%s"', $value));
  }

  /**
   * Convert a string to an attribute mapping.
   *
   * @param string $value
   *   The string to convert.
   *
   * @return array
   *   Attributes mapping.
   */
  protected static function convertStringToAttributesMapping(string $value): array {
    $parse_html = '<div ' . $value . '></div>';
    $attributes = [];
    foreach (Html::load($parse_html)->getElementsByTagName('div') as $div) {
      foreach ($div->attributes as $attr) {
        $attributes[$attr->nodeName] = $attr->nodeValue;
      }
    }
    return $attributes;
  }

}
